/*
  Hangul Drawing routine used in X

  1996.9 Choi Jun-Ho(junker@vishnu.snu.ac.kr)
  Narae, CS Dept, SNU.
*/

#include <X11/Xlib.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include "fvwm/fvwmlib.h"

#define ABS(x)    ((x)>0?(x):-(x))
/* utils */
static int EUC_KR2WC();
static XFontStruct* NXQueryMatchingFont();

/* globals */
static NXFont* hfont;
Display* NX_Display=NULL;

/* NXHFontInit()
   
   Initialize Dynamic Font Loading Environment.
   Not load fonts actually.

   Recommended to all immediately after XOpenDisplay() calls.
   */
void NXHFontInit(display)
Display *display;
{
  if (display==NULL) return; /* cannot init */
  NX_Display=display;

  hfont=NULL;
  return;
}

/* NXHRegisterFontByName()

   Add <fontname> to font list.
   First font registered is used as default font if appropriate font
   doesn't exist.

   This call load font using XLoadQueryFont() calls to get ascent
   metrics.
   */
void NXHRegisterFontByName(fontname)
char* fontname;
{
  NXFont* backup;

  if (NX_Display==NULL) return; /* cannot do without display */

  /* add font to chain */
  backup=hfont;
  hfont=(NXFont*)malloc(sizeof(NXFont));
  hfont->fontname=strdup(fontname);
  hfont->font=XLoadQueryFont(NX_Display, fontname);
  hfont->ascent=hfont->font->ascent;
  hfont->fid=hfont->font->fid;
  hfont->next=backup;
  
#ifdef DEBUG
  fprintf(stderr, "Font %s added\n", fontname);
#endif

  return;
}

/* NXDrawString()

   Equivalent to XDrawString() 8bit char version.
   If Korean character appears, find appropriate Korean font and
   display on this.
   */
void NXDrawString(display, d, gc, x, y, str, len)
Display *display;
Drawable d;
GC gc;
int x, y;
char *str;
int len;
{
  int n;
  char *p;
  XChar2b buf[256];
  XFontStruct *efont, *hfont;
  Font efid;
  GC hgc;
  XGCValues gcv;

  /* create GC for hangul font */
  hgc=XCreateGC(display, d, 0, (XGCValues *) NULL);
  XCopyGC(display, gc, ~(0L), hgc);

  /* get english font info */
  XGetGCValues(display, gc, GCFont, &gcv);
  efid=gcv.font;
  efont=XQueryFont(display, efid);

  /* get hangul font according to english font */
  hfont=NXQueryMatchingFont(display, efont->ascent);

  XSetFont(display, hgc, hfont->fid);
  
  while (len > 0) {
    if (*str & 0x80) { /* hangul chars */
      for (p = str; len > 0 && (*p & 0x80); p += 2, len -= 2)
        ;
      n = EUC_KR2WC(str, buf, p-str);
      XDrawString16(display, d, hgc, x, y, buf, n);
      x+=XTextWidth16(hfont, buf, n);
    }
    else { /* english chars */
      for (p = str; len > 0 && !(*p & 0x80); p++, len--)
        ;
      XDrawString(display, d, gc, x, y, str, p-str);
      x+=XTextWidth(efont, str, p-str);
    }
    str = p;
  }

  /* clean up */
  XFreeGC(display, hgc);
}

/* NXTextWidth()

   Equivalent to XTextWidth()
   */
int NXTextWidth(efont, str, len)
XFontStruct *efont;
char *str;
int len;
{
  int n;
  int width;
  char *p;
  XChar2b buf[256];
  XFontStruct *hfont;

  /* get hangul font according to english font */
  hfont=NXQueryMatchingFont(NX_Display, efont->ascent);

  width=0;
  while (len > 0) {
    if (*str & 0x80) { /* hangul chars */
      for (p = str; len > 0 && (*p & 0x80); p += 2, len -= 2)
        ;
      n = EUC_KR2WC(str, buf, p-str);
#ifdef DEBUG
      fprintf(stderr, "(hfont, buf, n)=(%p, %s, %d)\n", hfont, buf, n);
#endif
      width+=XTextWidth16(hfont, buf, n);
    }
    else { /* english chars */
      for (p = str; len > 0 && !(*p & 0x80); p++, len--)
        ;
      width+=XTextWidth(efont, str, p-str);
    }
    str = p;
  }
  
  return width;
}

/* NXHFontDestory()

   Clear all loading fonts.
   Recommended to use immediately before XCloseDisplay()
   */
void NXHFontDestroy()
{
  NXFont* backup;

  if (hfont==NULL) return;

  backup=hfont;
  while(backup!=NULL){
#ifdef DEBUG
    fprintf(stderr, "Free Font %s\n", backup->fontname);
#endif
    XFreeFont(NX_Display, backup->font);
    free(backup->fontname);
    backup=backup->next;
  }
  
  return;
}


/****** utils *******/

/* EUC_KR2WC()
   
   convert EUC-KR(KSC-5601) char string to XChar2b struct.
   corresfonding XLFD *-KSC5601.1987-0
   */
int EUC_KR2WC(str, buf, n)
char* str;
XChar2b* buf;
int n;
{
  int i, j;
  
  for (i = 0, j = 0; i < n;) {
    buf[j].byte1 = str[i] - 0x80;
    buf[j].byte2 = str[i+1] - 0x80;
    i += 2, j++;
  }
  return j;
}

/* NXQueryMatchingFont()

   following to <ascent>, find Korean font and returns XFontStruct*.
   */
XFontStruct* NXQueryMatchingFont(display, asc)
Display* display;
int asc;
{
  NXFont *curfont, *selfont;
  int min=255;

#ifdef DEBUG
  fprintf(stderr, "NXGetFID: Ascent %d..", asc);
#endif

  /* select best approxicated size font */
  curfont=hfont;
  selfont=NULL;
  while(curfont!=NULL){
#ifdef DEBUG
    fprintf(stderr, "%d..", curfont->ascent);
#endif
    if (ABS((asc-1)-curfont->ascent)<=min){
      min=ABS(asc-curfont->ascent);
      selfont=curfont;
#ifdef DEBUG
      fprintf(stderr, "(%d,%d)M%d", curfont->ascent, asc, min);
#endif
    }
    curfont=curfont->next;
  }

  /* error */
  if (selfont==NULL)
    return NULL;

#ifdef DEBUG
  fprintf(stderr, "Font %s selected\n", selfont->fontname);
#endif
  
  return selfont->font;
}
